/*jslint nomen: true, maxlen: 200, indent: 2, maxerr: 100*/
/*global window, document, URL, rJS, RSVP, jIO */

(function (window, document, rJS, RSVP, jIO) {
  "use strict";
  var render_object;

  function decodeJsonPointer(_str) {
    // https://tools.ietf.org/html/rfc6901#section-5
    return _str.replace(/~1/g, '/').replace(/~0/g, '~');
  }

  function encodeJsonPointer(_str) {
    // https://tools.ietf.org/html/rfc6901#section-5
    return _str.replace(/~/g, '~0').replace(/\//g, '~1');
  }

  function getBaseUrl(g, path) {
    var base_url,
      key,
      map = g.props.schema_map,
      max_len = 0;
    if (!path) {
      return;
    }
    for (key in map) {
      if (map.hasOwnProperty(key) &&
          path.startsWith(key) &&
          key.length > max_len) {
        base_url = map[key];
        max_len = key.length;
      }
    }
    return base_url;
  }

  function downloadJSON(url) {
    return RSVP.Queue()
      .push(function () {
        return jIO.util.ajax({
          url: url,
          dataType: "json"
        });
      })
      .push(function (evt) {
        return evt.target.response;
      });
  }

  function resolveLocalReference(schema, ref) {
    // 2 here is for #/
    var i, ref_path = ref.substr(2, ref.length),
      parts = ref_path.split("/");
    if (parts.length === 1 && parts[0] === "") {
      // It was uses #/ to reference the entire json so just return it.
      return schema;
    }
    for (i = 0; i < parts.length; i += 1) {
      schema = schema[decodeJsonPointer(parts[i])];
    }
    return schema;
  }

  function loadJSONSchema(g, url, path) {
    if (g.props.toplevel) {
      return g.loadJSONSchema(url, path);
    }
    path = g.element.getAttribute('data-json-parent') +
      encodeJsonPointer(g.element.getAttribute('data-json-property-name')) +
      path;
    return g.loadJSONSchemaParent(url, path);
  }

  function getDocumentType(doc) {
    if (doc instanceof Array) {
      return "array";
    }
    return typeof doc;
  }

  function getDocumentSchema(doc) {
    var type = getDocumentType(doc),
      schema = {
        type: type
      };
    if (type === "array") {
      schema.maxItems = 0;
    } else if (type === "object") {
      schema.additionalProperties = false;
    } else {
      schema.readOnly = true;
    }
    return schema;
  }

  function render_selection(json_field, default_value) {
    var input = document.createElement("select"),
      option = document.createElement("option"),
      option_index,
      optionz;
    input.size = 1;
    option.value = "";
    if (default_value === undefined) {
      option.selected = "selected";
    }
    input.appendChild(option);
    for (option_index in json_field['enum']) {
      if (json_field['enum'].hasOwnProperty(option_index)) {
        optionz = document.createElement("option");
        optionz.value = json_field['enum'][option_index];
        optionz.textContent = json_field['enum'][option_index];
        if (json_field['enum'][option_index] === default_value) {
          optionz.selected = "selected";
        }
        input.appendChild(optionz);
      }
    }
    return input;
  }

  function render_textarea(json_field, default_value, data_format) {
    var input = document.createElement("textarea");
    if (default_value !== undefined) {
      if (default_value instanceof Array) {
        input.value = default_value.join("\n");
      } else {
        input.value = default_value;
      }
    }
    input["data-format"] = data_format;
    return input;
  }

  function addSubForm(options) {
    var element = options.element,
      g = options.gadget,
      property_name,
      parent_path = options.path || element.name,
      scope,
      input_element;

    scope = parent_path + Math.random().toString(36).substr(2, 9);
    if (options.parent_type !== "array") {
      property_name = options.property_name;
      if (!property_name) {
        input_element = element.parentNode.querySelector("input[type='text']");
        property_name = input_element.value;
      }
      if (!property_name) {
        // XXX notify user
        // you can't create property without property_name
        return RSVP.Queue();
      }
      if (g.props.objects[parent_path].hasOwnProperty(property_name) && g.props.objects[parent_path][property_name] !== "") {
        // XXX notify user
        // you can't create property with existed property_name
        return RSVP.Queue();
      }
      if (input_element) {
        input_element.value = "";
      }
    }

    return g.declareGadget('gadget_json_generated_form.html', {scope: scope})
      .push(function (form_gadget) {
        form_gadget.element.setAttribute("data-json-parent", parent_path);
        form_gadget.element.setAttribute("data-gadget-parent-scope",
          g.element.getAttribute("data-gadget-scope"));
        if (options.parent_type === "array") {
          g.props.arrays[parent_path] = 1;
        } else {
          g.props.objects[parent_path][property_name] = scope;
          form_gadget.element.setAttribute("data-json-property-name", property_name);
        }
        return form_gadget.renderForm({
          schema: options.schema_part,
          document: options.default_dict,
          display_label: options.parent_type !== "array",
          scope: scope
        });
      });
  }

  function render_array(gadget, json_field, default_array, root, path) {
    var queue = RSVP.Queue(),
      div,
      div_input,
      input,
      item_schema,
      i,
      maxItems = json_field.maxItems;
    div = document.createElement("div");
    div.setAttribute("class", "subfield");
    div.title = json_field.description;

    div_input = document.createElement("div");
    div_input.setAttribute("class", "input");

    function element_append(child) {
      if (child) {
        div_input.appendChild(child);
      }
    }

    if (maxItems === undefined || default_array.length < maxItems) {
      item_schema = json_field.items;

      input = document.createElement("button");
      input.setAttribute("class",
        "ui-shadow-inset ui-btn ui-btn-inline ui-corner-all" +
        " ui-btn-icon-notext ui-icon-btn ui-icon-plus ui-input-btn");
      input.type = "button";
      input.name = path;
      gadget.props.add_buttons.push({
        element: input,
        event: function () {
          return addSubForm({
            gadget: gadget,
            parent_type: 'array',
            element: input,
            schema_part: item_schema
          })
            .push(element_append);
        }
      });
    } else {
      input = document.createElement("div");
      input.setAttribute("class", "input");
    }

    div_input.appendChild(input);
    div.appendChild(div_input);

    if (default_array) {
      for (i = 0; i < default_array.length; i = i + 1) {
        queue
          .push(
            addSubForm.bind(gadget, {
              gadget: gadget,
              parent_type: 'array',
              path: path,
              element: input,
              schema_part: item_schema,
              default_dict: default_array[i]
            })
          )
          .push(element_append);
      }
    }
    root.appendChild(div);
    // XXX add failback rendering if default_array not array
    // input = render_textarea(json_field, default_value, "array");
    return queue;
  }

  function render_field(gadget, key, path, json_field, default_value, root, editable_label) {
    var div,
      label,
      div_input,
      span_info,
      span_error,
      input,
      first_path,
      queue = RSVP.Queue();

    if (path && key) {
      first_path = path + encodeJsonPointer(key);
    } else {
      first_path = "";
    }

    if (json_field === undefined) {
      json_field = getDocumentSchema(default_value);
    }

    if (json_field.$ref !== undefined) {
      return loadJSONSchema(gadget, json_field.$ref, first_path)
        .push(function (schema_part) {
          return render_field(gadget, key, path, schema_part, default_value, root);
        });
    }

    if (json_field.type === undefined && default_value) {
      json_field.type = getDocumentType(default_value);
    }

    // XXX bad peace of code
    // we can not change schema
    // but our schema in slapos bad
    if (!json_field.type) {
      if (json_field.properties &&
          json_field.required &&
          json_field.required.length > 0) {
        json_field.type = "object";
      }
    }

    div = document.createElement("div");
    div.setAttribute("class", "subfield");
    div.title = json_field.description;
    // if (key && !first_path) {
    if (false) {
      // XXX;
      label = document.createElement("input");
      label.value = key;
      gadget.props.property_name_edit = label;
    } else {
      label = document.createElement("label");
      label.textContent = json_field.title || key;
    }
    div.appendChild(label);
    div_input = document.createElement("div");
    div_input.setAttribute("class", "input");

    if (json_field.enum !== undefined) {
      input = render_selection(json_field, default_value);
    }

    if (json_field.type === "boolean") {
      if (default_value === "true") {
        default_value = true;
      }
      if (default_value === "false") {
        default_value = false;
      }
      input = render_selection({
        type: "boolean",
        enum: [true, false]
      }, default_value);
    }

    if (!input && ["string", "integer", "number"].indexOf(json_field.type) >= 0) {
      if (json_field.contentMediaType === "text/plain") {
        input = render_textarea(json_field, default_value, "string");
      } else {
        input = document.createElement("input");
        if (default_value !== undefined) {
          input.value = default_value;
        }

        if (json_field.type === "integer") {
          input.type = "number";
        } else {
          input.type = "text";
        }
      }
    }

    if (json_field.type === "array") {
      queue = render_array(
        gadget,
        json_field,
        default_value,
        div_input,
        first_path + '/'
      );
      div.setAttribute("data-json-path", first_path + '/');
      div.setAttribute("data-json-type", json_field.type);
    }

    if (json_field.type === "object") {
      queue
        .push(function () {
          return render_object(
            gadget,
            json_field,
            default_value,
            div_input,
            first_path + '/'
          );
        });
      div.setAttribute("data-json-path", first_path + '/');
    }

    if (input) {
      // object and array excluded from
      // gadget.props.inputs not contain values
      gadget.props.inputs.push(input);
      input.name = first_path;
      input.setAttribute("class", "slapos-parameter");
      div_input.appendChild(input);
    }

    if (json_field.info !== undefined) {
      span_info = document.createElement("span");
      span_info.textContent = json_field.info;
      div_input.appendChild(span_info);
    }
    span_error = document.createElement("span");
    span_error.setAttribute("class", "error");
    div_input.appendChild(span_error);
    div.appendChild(div_input);

    return queue
      .push(function () {
        root.appendChild(div);
        return div;
      });
  }

  render_object = function (g, json_field, default_dict, root, path) {
    var additionalProperties,
      key,
      required = json_field.required || [],
      used_properties = {},
      queue = RSVP.Queue(),
      scope = "property_add_" + Math.random().toString(36).substr(2, 9),
      selector = {
        event: function () {
          return g.getDeclaredGadget(scope)
            .push(function (z) {
              return z.getContent()
                .push(function (value) {
                  var property_name = value[scope];
                  used_properties[property_name] = "";
                  return addSubForm({
                    gadget: g,
                    property_name: property_name,
                    path: path,
                    schema_part: json_field.properties[property_name]
                  });
                })
                .push(function (element) {
                  z.element.parentNode.insertBefore(element, z.element);
                });
            })
            .push(selector.rerender);
        },
        rerender: function () {
          return g.getDeclaredGadget(scope)
            .push(function (g) {
              var property_name,
                item_list = [["add property", "add property"]];
              for (property_name in json_field.properties) {
                if (json_field.properties.hasOwnProperty(property_name) &&
                    !used_properties.hasOwnProperty(property_name)) {
                  item_list.push([property_name, property_name]);
                }
              }
              return g.render({
                name: scope,
                editable: true,
                hidden: item_list.length <= 1,
                value: item_list[0][1],
                item_list: item_list
              })
              //not need if gadget_html5_select.render return element
                .push(function () {
                  return g.element;
                });
            });
        }
      };

    g.props.add_property_selections[scope] = selector;
    g.props.objects[path] = used_properties;

    function root_append(child) {
      root.appendChild(child);
    }

    function addAdditional(schema) {
      var div,
        div_input,
        input,
        property_name;

      div = document.createElement("div");
      div.setAttribute("class", "subfield");
      div.title = json_field.description;

      div_input = document.createElement("div");
      div_input.setAttribute("class", "input");

      function element_append(child) {
        if (child) {
          // insert additionalProperty before selector
          selector.element.parentNode.insertBefore(child, selector.element);
        }
      }

      input = document.createElement("input");
      input.type = "text";
      div_input.appendChild(input);

      input = document.createElement("button");
      input.setAttribute("class",
        "ui-shadow-inset ui-btn ui-btn-inline ui-corner-all" +
        " ui-btn-icon-notext ui-icon-btn ui-icon-plus ui-input-btn");
      input.type = "button";
      input.name = path;
      g.props.add_buttons.push({
        element: input,
        event: function () {
          return addSubForm({
            gadget: g,
            element: input,
            schema_part: schema
          })
            .push(element_append);
        }
      });
      div_input.appendChild(input);
      div.appendChild(div_input);


      for (property_name in default_dict) {
        if (default_dict.hasOwnProperty(property_name) && !used_properties.hasOwnProperty(property_name)) {
          used_properties[property_name] = "";
          queue
            .push(
              addSubForm.bind(g, {
                gadget: g,
                property_name: property_name,
                path: path,
                element: input,
                schema_part: schema,
                default_dict: default_dict[property_name]
              })
            )
            .push(element_append);
        }
      }
      queue.push(function () {
        root.appendChild(div);
      });
    }

    if (default_dict === undefined) {
      default_dict = {};
    }

    for (key in json_field.properties) {
      if (json_field.properties.hasOwnProperty(key)) {
        if (required.indexOf(key) >= 0) {
          used_properties[key] = false;
          if (json_field.properties[key].default !== undefined) {
            json_field.properties[key].info = '(default = ' + json_field.properties[key].default + ')';
          }
          queue
            .push(render_field.bind(g, g, key, path,
                json_field.properties[key], default_dict[key], root)
              );
        } else if (default_dict.hasOwnProperty(key)) {
          used_properties[key] = "";
          queue
            .push(
              addSubForm.bind(g, {
                gadget: g,
                property_name: key,
                path: path,
                schema_part: json_field.properties[key],
                default_dict: default_dict[key]
              })
            )
            .push(root_append);
        }
      }
    }

    queue
      .push(function () {
        return g.declareGadget("gadget_html5_select.html", {scope: scope});
      })
      .push(selector.rerender)
      .push(function (element) {
        selector.element = element;
        return root_append(element);
      });

    if (json_field.patternProperties !== undefined) {
      if (json_field.patternProperties['.*'] !== undefined) {
        addAdditional(json_field.patternProperties['.*']);
      }
    }

    if (json_field.additionalProperties === undefined) {
      additionalProperties = true;
    } else {
      additionalProperties = json_field.additionalProperties;
    }
    if (getDocumentType(additionalProperties) === "object") {
      addAdditional(additionalProperties);
    }

    for (key in default_dict) {
      if (default_dict.hasOwnProperty(key)) {
        if (!used_properties.hasOwnProperty(key)) {
          queue
            .push(render_field.bind(g, g, key, path, undefined, default_dict[key], root));
        }
      }
    }
    return queue;
  };

  function getFormValuesAsJSONDict(g) {
    var multi_level_dict = {"": {}},
      scope,
      options = g.props,
      array,
      path,
      key,
      i,
      len,
      queue = RSVP.Queue();

    function convertOnMultiLevel(d, key, value) {
      var ii,
        kk,
        key_list = key.split("/");
      for (ii = 0; ii < key_list.length; ii += 1) {
        kk = decodeJsonPointer(key_list[ii]);
        if (ii === key_list.length - 1) {
          if (value !== undefined) {
            d[kk] = value;
          } else {
            return d[kk];
          }
        } else {
          if (!d.hasOwnProperty(kk)) {
            d[kk] = {};
          }
          d = d[kk];
        }
      }
    }

    function recursiveGetContent(scope, path) {
      queue
        .push(function () {
          return g.getDeclaredGadget(scope);
        })
        .push(function (gadget) {
          return gadget.getContent();
        })
        .push(function (jdict) {
          convertOnMultiLevel(multi_level_dict, path, jdict);
        });
    }

    function getContentAndPushArray(scope, parent_path) {
      queue
        .push(function () {
          return g.getDeclaredGadget(scope);
        })
        .push(function (gadget) {
          return gadget.getContent();
        })
        .push(function (jdict) {
          var arr = convertOnMultiLevel(multi_level_dict, parent_path);
          if (!(arr instanceof Array)) {
            arr = [];
            convertOnMultiLevel(multi_level_dict, parent_path, arr);
          }
          arr.push(jdict);
        });
    }

    for (path in options.arrays) {
      if (options.arrays.hasOwnProperty(path)) {
        array = g.element.querySelector("div[data-json-path='" + path + "']")
          .querySelectorAll("div[data-json-parent='" + path + "']" +
            "[data-gadget-parent-scope='" + g.element.getAttribute("data-gadget-scope") + "']");
        len = array.length;
        for (i = 0; i < len; i = i + 1) {
          getContentAndPushArray(
            array[i].getAttribute('data-gadget-scope'),
            // slice remove concluding '/'
            path.slice(0, -1)
          );
        }
      }
    }

    for (path in options.objects) {
      if (options.objects.hasOwnProperty(path)) {
        for (key in options.objects[path]) {
          if (options.objects[path].hasOwnProperty(key)) {
            scope = options.objects[path][key];
            if (scope) {
              recursiveGetContent(scope, path + encodeJsonPointer(key));
            }
          }
        }
      }
    }

    return queue
      .push(function () {
        var json_dict = {},
          k;
        g.props.inputs.forEach(function (input) {
          if (input.name === "" || input.value !== "") {
            if (input.type === 'number') {
              json_dict[input.name] = parseInt(input.value, 10);
            } else if (input.value === "true") {
              json_dict[input.name] = true;
            } else if (input.value === "false") {
              json_dict[input.name] = false;
            } else if (input.tagName === "TEXTAREA") {
              if (input["data-format"] === "string") {
                json_dict[input.name] = input.value;
              } else {
                json_dict[input.name] = input.value.split('\n');
              }
            } else {
              json_dict[input.name] = input.value;
            }
          }
        });
        for (k in json_dict) {
          if (json_dict.hasOwnProperty(k)) {
            convertOnMultiLevel(multi_level_dict, k, json_dict[k]);
          }
        }
        return multi_level_dict[""];
      });
  }

  function getSubGadgetElement(g, scope) {
    return g.element.querySelector("div[data-gadget-scope='" + scope + "']");
  }

  rJS(window)
    .ready(function () {
      var g = this;
      g.props = {};
      g.options = {};
    })
    .declareAcquiredMethod("notifyValid", "notifyValid")
    .declareAcquiredMethod("notifyInvalid", "notifyInvalid")

    .declareAcquiredMethod("renameChildrenParent", "renameChildren")
    .allowPublicAcquisition("renameChildren", function (opt_arr, scope) {
      var property_name,
        objects = this.props.objects,
        new_name = opt_arr[0],
        element = getSubGadgetElement(this, scope),
        parent = element.getAttribute('data-json-parent');
      if (objects.hasOwnProperty(parent)) {
        parent = objects[parent];
        if (parent.hasOwnProperty(new_name)) {
          throw new Error("property already exist");
        }
        // XXX validate property if property pattern
        for (property_name in parent) {
          if (parent.hasOwnProperty(property_name) && parent[property_name] === scope) {
            delete parent[property_name];
            parent[new_name] = scope;
            return new_name;
          }
        }
        throw new Error("gadget not found for renaming");
      }
    })
    .declareMethod("rename", function (new_name, event) {
      var g = this,
        name = g.element.getAttribute('data-json-property-name');
      return this.renameChildrenParent(new_name)
        .push(function () {
          return g.element.setAttribute('data-json-property-name', new_name);
        })
        .push(undefined, function (error) {
          // XXX notify user
          event.srcElement.value = name;
          event.srcElement.focus();
        });
    })
    .declareAcquiredMethod("selfRemove", "deleteChildren")
    .allowPublicAcquisition("deleteChildren", function (arr, scope) {
      var g = this,
        key,
        objects = this.props.objects,
        element = getSubGadgetElement(g, scope),
        parent = element.getAttribute("data-json-parent"),
        tasks = [];
      if (objects.hasOwnProperty(parent)) {
        parent = objects[parent];
        for (key in parent) {
          if (parent.hasOwnProperty(key) && parent[key] === scope) {
            delete parent[key];
          }
        }
      }
      element.parentNode.removeChild(element);
      for (key in g.props.add_property_selections) {
        if (g.props.add_property_selections.hasOwnProperty(key)) {
          tasks.push(g.props
            .add_property_selections[key]
            .rerender());
        }
      }
      return RSVP.All(tasks);
    })

    .declareAcquiredMethod("processValidationParent", "processValidation")
    .allowPublicAcquisition("processValidation", function (json_dict) {
      return this.processValidation(undefined, json_dict);
    })
    .declareMethod('processValidation', function (schema_url, json_dict) {
      var g = this;
      if (!schema_url) {
        if (g.options.schema_url) {
          schema_url = g.options.schema_url;
        } else {
          return g.processValidationParent(json_dict);
        }
      }
      if (!json_dict) {
        json_dict = getFormValuesAsJSONDict(g);
      } else {
        json_dict = RSVP.Queue()
          .push(function () {
            return json_dict;
          });
      }
      return RSVP.Queue()
        .push(function () {
          return RSVP.all([
            g.getDeclaredGadget('loadschema'),
            json_dict
          ]);
        })
        .push(function (ret) {
          var loadschema_gadget = ret[0],
            json_d = ret[1];
          return loadschema_gadget.validateJSON(schema_url, json_d);
        })
        .push(function (validation) {
          var index,
            div,
            field_name;

          g.element.querySelectorAll("span.error").forEach(function (span) {
            span.textContent = "";
          });

          g.element.querySelectorAll("div.error-input").forEach(function (div) {
            div.setAttribute("class", "");
          });
          if (validation.valid) {
            return "VALID";
          }
          for (index in validation.errors) {
            if (validation.errors.hasOwnProperty(index)) {
              field_name = validation.errors[index].dataPath;
              div = g.element.querySelector(".slapos-parameter[name='" + field_name  + "']").parentNode;
              div.setAttribute("class", "slapos-parameter error-input");
              div.querySelector("span.error").textContent = validation.errors[index].message;
            }
          }

          for (index in validation.missing) {
            if (validation.missing.hasOwnProperty(index)) {
              field_name = validation.missing[index].dataPath;
              div = g.element.querySelector('.slapos-parameter[name=' + field_name  + "']").parentNode;
              div.setAttribute("class", "error-input");
              div.querySelector("span.error").textContent = validation.missing[index].message;
            }
          }
          return "ERROR";
        });
    })

    .allowPublicAcquisition("notifyValid", function (arr, sub_scope) {
      return true;
    })
    .allowPublicAcquisition("notifyChange", function (arr, sub_scope) {
      var g = this,
        opt = arr[0];
      if (opt.type === "change") {
        return g.props
          .add_property_selections[sub_scope]
          .event();
      }
    })
    .declareMethod('renderForm', function (options) {
      var g = this,
        property_name = g.element.getAttribute('data-json-property-name'),
        schema = options.schema,
        delete_button,
        root;
      g.props.inputs = [];
      g.props.add_buttons = [];
      g.props.add_property_selections = {};
      g.props.arrays = {};
      g.props.objects = {};
      g.props.path = options.path; // self gadget scope
      if (!property_name || !options.display_label) {
        property_name = "";
      }
      root = g.element.querySelector('[data-json-path="/"]');
      if (!root) {
        root = g.element;
      }
      while (root.firstChild) {
        root.removeChild(root.firstChild);
      }
      if (!g.props.toplevel) {
        delete_button = document.createElement("button");
        delete_button.setAttribute("class",
          "ui-shadow-inset ui-btn ui-btn-inline ui-corner-all" +
          " ui-btn-icon-notext ui-icon-btn ui-icon-trash-o ui-input-btn");
        delete_button.type = "button";
        delete_button.name = options.path;
        g.props.delete_button = delete_button;
        root.appendChild(delete_button);
      }
      return render_field(g, property_name, "", schema, options.document, root)
        .push(function () {
          g.listenEvents();
          return g.element;
        });
    })

    .declareMethod("loadJSONSchema", function (url, path) {
      var g = this,
        protocol,
        download_url,
        base_url,
        hash;
      // XXX need use $id
      base_url = getBaseUrl(g, path);
      if (!base_url) {
        // allow relative link
        base_url = window.location;
      }
      if (!path) {
        path = "/";
      }
      url = new URL(url, base_url);
      protocol = url.protocol;
      if (protocol === "http:" || protocol === "https:") {
        if (window.location.protocol !==  protocol) {
          throw new Error("You cannot mixed http and https calls");
        }
      }
      download_url = url.origin + url.pathname;
      hash = url.hash;
      url = url.href;
      return RSVP.Queue()
        .push(function () {
          if (g.props.schema_cache.hasOwnProperty(download_url)) {
            return g.props.schema_cache[download_url];
          }
          return downloadJSON(download_url)
            .push(function (json) {
              g.props.schema_cache[download_url] = json;
              return json;
            });
        })
        .push(function (json) {
          g.props.schema_map[path] = url;
          return resolveLocalReference(json, hash);
        });
    })
    .declareAcquiredMethod("loadJSONSchemaParent", "loadJSONSchemaTop")
    .allowPublicAcquisition("loadJSONSchemaTop", function (arr) {
      return loadJSONSchema(this, arr[0], arr[1]);
    })
    .declareMethod('render', function (options) {
      var g = this,
        queue;
      g.props.toplevel = true;
      // contain map of current normalized schema
      // json pointer and corresponding url
      // it's need for schema uri computation
      g.props.schema_map = {};
      g.props.schema_cache = {};
      g.options = options;
      if (!options.value) {
        options.value = {};
      }
      if (options.schema) {
        queue = RSVP.Queue()
          .push(function () {
            return options.schema;
          });
      } else {
        queue = loadJSONSchema(g, options.schema_url);
      }
      return queue
        .push(function (schema) {
          g.options.schema = schema;
          return g.renderForm({
            schema: schema,
            document: options.value
          });
        })
        .push(function () {
          return g;
        });
    })

    .onEvent('click', function (evt) {
      if (evt.target === this.props.delete_button) {
        return this.selfRemove(evt);
      }

      var button_list = this.props.add_buttons,
        i;
      for (i = 0; i < button_list.length; i = i + 1) {
        if (evt.target === button_list[i].element) {
          return button_list[i].event(evt);
        }
      }
    })

    .onEvent('change', function (evt) {
      if (evt.target === this.props.property_name_edit) {
        return this.rename(this.props.property_name_edit.value, evt);
      }

      var field_list = this.props.inputs,
        i;
      for (i = 0; i < field_list.length; i = i + 1) {
        if (evt.target === field_list[i].element) {
          return this.processValidation.bind(this, this.options.schema_url, undefined)(evt);
        }
      }
    })
    .declareJob('listenEvents', function () {
      // XXX Disable
      return;
    })

    .declareMethod('getContent', function () {
      var g = this;
      return getFormValuesAsJSONDict(g);
      // return g.processValidation(g.options.schema_url, json_dict)
      //   .push(function (status) {
      //     return {
      //       value: json_dict,
      //       status: status
      //     };
      //   });
    });

}(window, document, rJS, RSVP, jIO));